import { __read, __assign, __spread, __decorate, __param, __extends, __awaiter, __generator } from 'tslib';
import { Validators, NG_VALUE_ACCESSOR } from '@angular/forms';
import { circularDeepEqual } from 'fast-equals';
import { CommonModule } from '@angular/common';
import { InjectionToken, EventEmitter, ElementRef, Renderer2, Inject, Input, Output, Directive, HostListener, Optional, ContentChildren, ChangeDetectorRef, NgModule } from '@angular/core';
import { asyncScheduler, Subject, BehaviorSubject, Subscription, merge, combineLatest, ReplaySubject } from 'rxjs';
import { throttleTime, switchMap, map, startWith, first } from 'rxjs/operators';

/**
 * Maps an object of control states to a value given by the `mapFunc`.
 *
 * @param controls Object which contains keys and associated control update.
 * @param mapFunc Mapper function to convert the update.
 *
 * @example
 * // Mapping of:
 * const controls = {
 *  firstControl: {
 *    value: 'firstValue',
 *    ...
 *  },
 *  secondControl: {
 *    value: 'secondValue',
 *    ...
 *  },
 * }
 *
 * // With the function:
 * const mapFunc = (control) => control.value
 *
 * // Will result in:
 * const result = {
 *  firstControl: 'firstValue',
 *  secondControl: 'secondValue'
 * }
 */
function mapFormGroupControlStates(controls, mapFunc) {
    return mapFormControls(controls, mapFunc);
}
/**
 * Maps an object of control summaries to a value given by the `mapFunc`.
 *
 * @param controls Object which contains keys and associated control summaries.
 * @param mapFunc Mapper function to convert the update.
 *
 * @example
 * // Mapping of:
 * const controls = {
 *  firstControl: {
 *    value: 'firstValue',
 *    ...
 *  },
 *  secondControl: {
 *    value: 'secondValue',
 *    ...
 *  },
 * }
 *
 * // With the function:
 * const mapFunc = (control) => control.value
 *
 * // Will result in:
 * const result = {
 *  firstControl: 'firstValue',
 *  secondControl: 'secondValue'
 * }
 */
function mapFormGroupControlSummaries(controls, mapFunc) {
    return mapFormControls(controls, mapFunc);
}
/**
 * Maps an object of control updates to a value given by the `mapFunc`.
 *
 * @param controls Object which contains keys and associated control updates.
 * @param mapFunc Mapper function to convert the update.
 *
 * @example
 * // Mapping of:
 * const controls = {
 *  firstControl: {
 *    value: 'firstValue',
 *    ...
 *  },
 *  secondControl: {
 *    value: 'secondValue',
 *    ...
 *  },
 * }
 *
 * // With the function:
 * const mapFunc = (control) => control.value
 *
 * // Will result in:
 * const result = {
 *  firstControl: 'firstValue',
 *  secondControl: 'secondValue'
 * }
 */
function mapFormGroupControlUpdates(controls, mapFunc) {
    return mapFormControls(controls, mapFunc);
}
function mapFormControls(controls, mapFunc) {
    var result = Object.entries(controls)
        .map(function (_a) {
        var _b;
        var _c = __read(_a, 2), key = _c[0], control = _c[1];
        return (_b = {},
            _b[key] = mapFunc(control, key),
            _b);
    })
        .reduce(function (ctrl1, ctrl2) { return (__assign(__assign({}, ctrl1), ctrl2)); }, {});
    return result;
}
/**
 * Converts one or multiple Angular validators for the usage with this framework.
 * Does **not** support asynchronous validators.
 * Uses Validators.compose from @angular/forms.
 *
 * Only a subset of properties is available for the Angular validators:
 * `value, dirty, pristine, touched, untouched, enabled, disabled`
 * @param fn Angular validator function..
 */
function validatorOf() {
    var fn = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        fn[_i] = arguments[_i];
    }
    var composed = Validators.compose(fn);
    return function (control) {
        return composed({
            value: control.value,
            dirty: !control.pristine,
            pristine: control.pristine,
            touched: !control.untouched,
            untouched: control.untouched,
            enabled: !control.disabled,
            disabled: control.disabled,
        });
    };
}
/**
 * Merges an array of `FormControlErrors`.
 * Returns null if all errors are null.
 * @param errors An array of `FormControlErrors` which will be merged into a single `FormControlErrors` value.
 */
function mergeFormControlErrors() {
    var errors = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        errors[_i] = arguments[_i];
    }
    return errors.reduce(function (e1, e2) {
        if (!e1 && !e2) {
            return null;
        }
        if (!e1) {
            return e2;
        }
        if (!e2) {
            return e1;
        }
        return __assign(__assign({}, e1), e2);
    }, null);
}
/**
 * Merges an array of errors.
 * The different child control errors will also be merged.
 * Returns null if all errors are null.
 * @param errors Multiple errors to be merged.
 *
 * @example
 * // Merge of:
 * {
 *   stringControl: {
 *     firstError: 'firstError'
 *   }
 * },
 * {
 *   stringControl: {
 *     secondError: 'secondError'
 *   }
 * }
 *
 * // Will result in:
 * {
 *   stringControl: {
 *     firstError: 'firstError',
 *     secondError: 'secondError'
 *   }
 * }
 *
 */
function mergeFormGroupErrors() {
    var errors = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        errors[_i] = arguments[_i];
    }
    return errors.reduce(function (group1, group2) {
        if (!group1 && !group2) {
            return null;
        }
        if (!group1) {
            return group2;
        }
        if (!group2) {
            return group1;
        }
        return __assign(__assign(__assign({}, group1), group2), Object.keys(group1)
            .filter(function (key1) { return Object.keys(group2).find(function (key2) { return key1 === key2; }); })
            .map(function (key) {
            var _a;
            return (_a = {},
                _a[key] = mergeFormControlErrors(group1[key], group2[key]),
                _a);
        })
            .reduce(function (e1, e2) { return (__assign(__assign({}, e1), e2)); }, {}));
    }, null);
}
/**
 * Merges an array of errors.
 * Errors at each index will also be merged.
 * Returns null if all errors are null.
 * @param errors Multiple errors to be merged.
 *
 * @example
 * // Merge of:
 * [
 *   {
 *     firstError: 'firstError'
 *   }
 * ],
 * [
 *   stringControl: {
 *     secondError: 'secondError'
 *   }
 * ]
 *
 * // Will result in:
 * [
 *   {
 *     firstError: 'firstError',
 *     secondError: 'secondError'
 *   }
 * ]
 *
 */
function mergeFormArrayErrors() {
    var errors = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        errors[_i] = arguments[_i];
    }
    var mergedErrors = errors
        .filter(Boolean)
        .reduce(function (arr1, arr2) { return (arr1.length >= arr2.length ? arr1 : arr2); }, [])
        .map(function (_, i) {
        return mergeFormControlErrors.apply(void 0, __spread(errors.map(function (arrayErrors) { return arrayErrors && arrayErrors[i]; })));
    });
    return mergedErrors.filter(Boolean).length > 0 ? mergedErrors : null;
}

/**
 * Returns a `FormControlState` which represents the input state with the applied update.
 * The returned value is a new constructed object, except when the update is invalid.
 *
 * @param control The `FormControlState` which the update will be applied to.
 * @param update The `FormControlUpdate` itself.
 */
function reduceFormControl(control, update) {
    if (!update) {
        return control;
    }
    return __assign(__assign({}, control), update);
}
/**
 * Returns a `FormGroupState` which represents the input state with the applied update.
 * The returned value is a new constructed object, except when the update is invalid.
 *
 * @param control The `FormGroupState` which the update will be applied to.
 * @param update The `FormGroupUpdate` itself.
 */
function reduceFormGroup(group, update) {
    if (!update) {
        return group;
    }
    return __assign(__assign(__assign({}, group), update), { controls: __assign(__assign({}, group.controls), (update.controls
            ? mapFormGroupControlUpdates(update.controls, function (control, key) { return (__assign(__assign({}, group.controls[key]), control)); })
            : {})) });
}
/**
 * Returns a `FormArrayState` which represents the input state with the applied update.
 * The returned value is a new constructed object, except when the update is invalid.
 *
 * @param control The `FormArrayState` which the update will be applied to.
 * @param update The `FormArrayUpdate` itself. Invalid values and empty objects inside the `update.controls` array will be ignored.
 */
function reduceFormArray(group, update) {
    if (!update) {
        return group;
    }
    return __assign(__assign(__assign({}, group), update), { controls: group.controls.map(function (control, index) {
            return update && update.controls ? reduceFormControl(control, update.controls[index]) : control;
        }) });
}

/**
 * Initializes a new `FormControlState`.
 * @see `FormControlInit`
 *
 * @example
 * initFormControl(['value', [validator]])
 * initFormControl({value: 'value', validators: [validator], disabled: true, ...})
 */
function initFormControl(initial) {
    return Array.isArray(initial)
        ? initFormControlFromTuple(initial)
        : initFormControlFromUpdate(initial);
}
/**
 * Initializes a new `FormGroupState`.
 * @param controls An object containing control initializations.
 * @param validators An array of GroupValidators which will be used for additional error detection.
 * @see `FormGroupInit`, `GroupValidator`
 *
 * @example
 * initFormGroup({
 *  firstControl: ['firstValue', [validator]],
 *  secondControl: {value: 'value', validators: [validator], disabled: true, ...}
 * })
 */
function initFormGroup(controls, validators) {
    if (validators === void 0) { validators = []; }
    return {
        controls: Object.entries(controls)
            .map(function (_a) {
            var _b;
            var _c = __read(_a, 2), key = _c[0], initial = _c[1];
            return (_b = {},
                _b[key] = initFormControl(initial),
                _b);
        })
            .reduce(function (ctrl1, ctrl2) { return (__assign(__assign({}, ctrl1), ctrl2)); }, {}),
        validators: validators,
    };
}
/**
 * Initializes a new `FormArrayState`.
 * @param controls An object containing control initializations.
 * @param validators An array of ArrayValidators which will be used for additional error detection.
 * @see `FormArrayInit`, `ArrayValidator`
 *
 * @example
 * initFormArray(
 *  [
 *      ['firstValue', [validator]],
 *      {value: 'value', validators: [validator], disabled: true, ...}
 *  ],
 *  [arrayValidator])
 */
function initFormArray(initial, validators) {
    if (validators === void 0) { validators = []; }
    return {
        controls: initial.map(function (init) { return initFormControl(init); }),
        validators: validators,
    };
}
function initFormControlFromTuple(_a) {
    var _b = __read(_a, 3), value = _b[0], _c = _b[1], validators = _c === void 0 ? [] : _c, _d = _b[2], disabled = _d === void 0 ? false : _d;
    return initFormControlFromUpdate({ value: value, validators: validators, disabled: disabled });
}
function initFormControlFromUpdate(initialUpdate) {
    return reduceFormControl({
        value: undefined,
        initialValue: initialUpdate.value,
        pristine: true,
        untouched: true,
        disabled: false,
        validators: [],
    }, initialUpdate);
}

/**
 * Resets a control back to the default values.
 * The following values will be ignored and stay the same:
 * - `initialValue`
 * - `validators`
 * - `disabled`
 *
 * @param control The `FormControlState` which should be used to create the reset.
 * @param initialValue Optional parameter for passing a new initial value.
 */
function resetFormControl(control, initialValue) {
    if (initialValue === void 0) { initialValue = control.initialValue; }
    return initFormControl([initialValue, control.validators, control.disabled]);
}
/**
 * Resets a group back to its default values.
 * Internally calls `resetFormControl` for every control.
 *
 * @see `resetFormControl`
 *
 * @param group The `FormGroupState` which should be used to create the reset.
 */
function resetFormGroup(group) {
    var controls = mapFormGroupControlStates(group.controls, function (control) {
        return resetFormControl(control);
    });
    return reduceFormGroup(group, { controls: controls });
}
/**
 * Resets an array back to its default values.
 * Internally calls `resetFormControl` for every control.
 *
 * @see `resetFormControl`
 *
 * @param array The `FormArrayState` which should be used to create the reset.
 */
function resetFormArray(array) {
    var controls = array.controls.map(function (control) { return resetFormControl(control); });
    return reduceFormArray(array, { controls: controls });
}

function getFormControlErrors(control) {
    return control.validators.map(function (validator) { return validator(control); });
}
function getFormGroupControlSummariesErrors(summaries) {
    var errors = mapFormGroupControlSummaries(summaries, function (summary) { return summary.errors; });
    Object.entries(errors)
        .filter(function (_a) {
        var _b = __read(_a, 2), _ = _b[0], value = _b[1];
        return value === null;
    })
        .forEach(function (_a) {
        var _b = __read(_a, 2), key = _b[0], _ = _b[1];
        return delete errors[key];
    });
    return Object.keys(errors).length ? errors : null;
}
function getFormArrayControlSummariesErrors(summaries) {
    var errors = summaries.map(function (summary) { return summary.errors; });
    return errors.filter(Boolean).length ? errors : null;
}
function getFormGroupErrors(group) {
    return mergeFormGroupErrors.apply(void 0, __spread(group.validators.map(function (validator) { return validator(group); })));
}
function getFormArrayErrors(array) {
    return mergeFormArrayErrors.apply(void 0, __spread(array.validators.map(function (validator) { return validator(array); })));
}
function getFormControlChanged(control) {
    return !circularDeepEqual(control.value, control.initialValue);
}
/**
 * Creates a `FormControlSummary` from the given `FormControlState`.
 * It is possible to add additional errors.
 *
 * @param control The `FormControlState` which is used to create the `FormControlSummary`.
 * @param additionalErrors An array of `FormControlErrors` which will be merged into the errors of the control.
 */
function getFormControlSummary(control) {
    var additionalErrors = [];
    for (var _i = 1; _i < arguments.length; _i++) {
        additionalErrors[_i - 1] = arguments[_i];
    }
    var errors = mergeFormControlErrors.apply(void 0, __spread(getFormControlErrors(control), additionalErrors));
    return __assign(__assign({}, control), { errors: errors, valid: errors === null, changed: getFormControlChanged(control) });
}
function getFormGroupControlSummaries(controls) {
    var additionalErrors = [];
    for (var _i = 1; _i < arguments.length; _i++) {
        additionalErrors[_i - 1] = arguments[_i];
    }
    var additionalError = mergeFormGroupErrors.apply(void 0, __spread(additionalErrors));
    return mapFormGroupControlStates(controls, function (control, key) {
        return getFormControlSummary(control, additionalError ? additionalError[key] : null);
    });
}
function getFormArrayControlSummaries(controls) {
    var additionalErrors = [];
    for (var _i = 1; _i < arguments.length; _i++) {
        additionalErrors[_i - 1] = arguments[_i];
    }
    var additionalError = mergeFormArrayErrors.apply(void 0, __spread(additionalErrors));
    return controls.map(function (control, i) {
        return getFormControlSummary(control, additionalError ? additionalError[i] : null);
    });
}
function getFormGroupPristine(group) {
    return Object.values(group.controls).every(function (control) { return control.pristine; });
}
function getFormArrayPristine(array) {
    return array.controls.every(function (control) { return control.pristine; });
}
function getFormGroupUntouched(group) {
    return Object.values(group.controls).every(function (control) { return control.untouched; });
}
function getFormGroupChanged(controls) {
    return Object.values(controls).some(function (control) { return control.changed; });
}
function getFormArrayChanged(controls) {
    return controls.some(function (control) { return control.changed; });
}
function getFormArrayUntouched(array) {
    return array.controls.every(function (control) { return control.untouched; });
}
function getFormArrayKeys(array) {
    return array.controls.map(function (_, i) { return i; });
}
/**
 * Creates a `FormGroupSummary` from the given `FormGroupState`.
 * It is possible to add additional errors.
 *
 * @param group The input `FormGroupState`. Used to create the `FormGroupSummary`.
 * @param additionalErrors An array of additional `FormGroupErrors`, which will be merged into the errors of the group and of each control.
 */
function getFormGroupSummary(group) {
    var additionalErrors = [];
    for (var _i = 1; _i < arguments.length; _i++) {
        additionalErrors[_i - 1] = arguments[_i];
    }
    var groupErrors = getFormGroupErrors(group);
    var summaries = getFormGroupControlSummaries.apply(void 0, __spread([group.controls,
        groupErrors], additionalErrors));
    var errors = getFormGroupControlSummariesErrors(summaries);
    return {
        controls: summaries,
        pristine: getFormGroupPristine(group),
        untouched: getFormGroupUntouched(group),
        errors: errors,
        valid: errors === null,
        changed: getFormGroupChanged(summaries),
        validators: group.validators,
    };
}
/**
 * Creates a `FormArraySummary` from the given `FormArrayState`.
 * It is possible to add additional errors.
 *
 * @param array The input `FormArrayState`. Used to create the `FormArraySummary`.
 * @param additionalErrors An array of additional `FormArrayErrors`, which will be merged into the errors of the array and of each control.
 */
function getFormArraySummary(array) {
    var additionalErrors = [];
    for (var _i = 1; _i < arguments.length; _i++) {
        additionalErrors[_i - 1] = arguments[_i];
    }
    var arrayErrors = getFormArrayErrors(array);
    var summaries = getFormArrayControlSummaries.apply(void 0, __spread([array.controls,
        arrayErrors], additionalErrors));
    var errors = getFormArrayControlSummariesErrors(summaries);
    return {
        controls: summaries,
        keys: getFormArrayKeys(array),
        pristine: getFormArrayPristine(array),
        untouched: getFormArrayUntouched(array),
        errors: errors,
        valid: errors === null,
        changed: getFormArrayChanged(summaries),
        validators: array.validators,
    };
}

var CONFIG_TOKEN = new InjectionToken('NGRX_CLEAN_FORMS_CONFIG');
var defaultConfig = {
    throttleTime: 20,
    distinctWritesOnly: true,
};
var throttle = function (config) {
    return throttleTime(config.throttleTime, asyncScheduler, {
        leading: true,
        trailing: true,
    });
};

var cssClasses = {
    invalid: 'ng-invalid',
    valid: 'ng-valid',
    pristine: 'ng-pristine',
    dirty: 'ng-dirty',
    touched: 'ng-touched',
    untouched: 'ng-untouched',
    changed: 'ng-changed',
    initial: 'ng-initial',
};
var CONTROL_DIRECTIVE_SELECTOR = "ngrxFormControl";
var AbstractControlDirective = /** @class */ (function () {
    function AbstractControlDirective(ref, r2, injectedConfig) {
        var _this = this;
        this.ref = ref;
        this.r2 = r2;
        this.injectedConfig = injectedConfig;
        this.controlUpdate = new EventEmitter(true);
        this.touched$ = new Subject();
        this.value$ = new Subject();
        this.config$ = new BehaviorSubject(this.injectedConfig);
        this.summarySubscription = new Subscription();
        this.config$
            .pipe(switchMap(function (config) {
            return merge(_this.touched$.pipe(throttle(config)), _this.value$.pipe(throttle(config)));
        }))
            .subscribe(function (value) { return _this.controlUpdate.emit(value); });
    }
    Object.defineProperty(AbstractControlDirective.prototype, "setControlSummary", {
        set: function (controlSummary) {
            if (!controlSummary) {
                return;
            }
            this.updateSummary(controlSummary);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(AbstractControlDirective.prototype, "setConfig", {
        set: function (inputConfig) {
            var config = __assign(__assign({}, this.injectedConfig), inputConfig);
            this.config$.next(config);
        },
        enumerable: true,
        configurable: true
    });
    AbstractControlDirective.prototype.isValueChanged = function (nextValue) {
        var disabled = !this.config$.value.distinctWritesOnly;
        var lastValue = this.getValue();
        return disabled || !circularDeepEqual(nextValue, lastValue);
    };
    AbstractControlDirective.prototype.emitTouched = function () {
        this.touched$.next({ untouched: false });
    };
    AbstractControlDirective.prototype.emitValue = function (value) {
        this.value$.next({ value: value, pristine: false });
    };
    AbstractControlDirective.prototype.updateSummary = function (summary) {
        if (this.isValueChanged(summary.value)) {
            this.setValue(summary.value);
        }
        this.setDisabled(summary.disabled);
        this.chooseClass(cssClasses.invalid, cssClasses.valid, summary.valid);
        this.chooseClass(cssClasses.dirty, cssClasses.pristine, summary.pristine);
        this.chooseClass(cssClasses.touched, cssClasses.untouched, summary.untouched);
        this.chooseClass(cssClasses.initial, cssClasses.changed, summary.changed);
    };
    AbstractControlDirective.prototype.chooseClass = function (class1, class2, chooseSecond) {
        if (chooseSecond) {
            this.r2.addClass(this.ref.nativeElement, class2);
            this.r2.removeClass(this.ref.nativeElement, class1);
        }
        else {
            this.r2.addClass(this.ref.nativeElement, class1);
            this.r2.removeClass(this.ref.nativeElement, class2);
        }
    };
    AbstractControlDirective.prototype.ngOnDestroy = function () {
        this.summarySubscription.unsubscribe();
        this.config$.complete();
        this.touched$.complete();
        this.value$.complete();
    };
    AbstractControlDirective.ctorParameters = function () { return [
        { type: ElementRef },
        { type: Renderer2 },
        { type: undefined, decorators: [{ type: Inject, args: [CONFIG_TOKEN,] }] }
    ]; };
    __decorate([
        Input(CONTROL_DIRECTIVE_SELECTOR)
    ], AbstractControlDirective.prototype, "controlKey", void 0);
    __decorate([
        Input('controlSummary')
    ], AbstractControlDirective.prototype, "setControlSummary", null);
    __decorate([
        Input('controlConfig')
    ], AbstractControlDirective.prototype, "setConfig", null);
    __decorate([
        Output()
    ], AbstractControlDirective.prototype, "controlUpdate", void 0);
    AbstractControlDirective = __decorate([
        Directive(),
        __param(2, Inject(CONFIG_TOKEN))
    ], AbstractControlDirective);
    return AbstractControlDirective;
}());

var RangeInputControlDirective = /** @class */ (function (_super) {
    __extends(RangeInputControlDirective, _super);
    function RangeInputControlDirective(ref, r2, config) {
        return _super.call(this, ref, r2, config) || this;
    }
    RangeInputControlDirective.prototype.onInput = function () {
        this.emitValue(this.getValue());
    };
    RangeInputControlDirective.prototype.onBlur = function () {
        this.emitTouched();
    };
    RangeInputControlDirective.prototype.getValue = function () {
        var value = this.ref.nativeElement.value;
        return value === '' ? null : parseFloat(value);
    };
    RangeInputControlDirective.prototype.setValue = function (value) {
        this.r2.setProperty(this.ref.nativeElement, 'value', value);
    };
    RangeInputControlDirective.prototype.setDisabled = function (disabled) {
        this.r2.setProperty(this.ref.nativeElement, 'disabled', disabled);
    };
    RangeInputControlDirective.ctorParameters = function () { return [
        { type: ElementRef },
        { type: Renderer2 },
        { type: undefined, decorators: [{ type: Inject, args: [CONFIG_TOKEN,] }] }
    ]; };
    __decorate([
        HostListener('input')
    ], RangeInputControlDirective.prototype, "onInput", null);
    __decorate([
        HostListener('blur')
    ], RangeInputControlDirective.prototype, "onBlur", null);
    RangeInputControlDirective = __decorate([
        Directive({
            selector: "input[type=\"range\"][" + CONTROL_DIRECTIVE_SELECTOR + "]",
        }),
        __param(2, Inject(CONFIG_TOKEN))
    ], RangeInputControlDirective);
    return RangeInputControlDirective;
}(AbstractControlDirective));

var RadioControlNotSupported = Error(
// tslint:disable-next-line: max-line-length
"Radio inputs are not yet supported. If you need to use one consider wrapping it in a Component which implements the ControlValueAccessor interface.");
var RadioInputControlDirective = /** @class */ (function (_super) {
    __extends(RadioInputControlDirective, _super);
    function RadioInputControlDirective(ref, r2, config) {
        var _this = _super.call(this, ref, r2, config) || this;
        throw RadioControlNotSupported;
        return _this;
    }
    RadioInputControlDirective.prototype.setValue = function (value) { };
    RadioInputControlDirective.prototype.getValue = function () {
        return null;
    };
    RadioInputControlDirective.ctorParameters = function () { return [
        { type: ElementRef },
        { type: Renderer2 },
        { type: undefined, decorators: [{ type: Inject, args: [CONFIG_TOKEN,] }] }
    ]; };
    RadioInputControlDirective = __decorate([
        Directive({
            selector: "input[type=\"radio\"][" + CONTROL_DIRECTIVE_SELECTOR + "]",
        }),
        __param(2, Inject(CONFIG_TOKEN))
    ], RadioInputControlDirective);
    return RadioInputControlDirective;
}(AbstractControlDirective));

var SelectControlNotSupported = Error(
// tslint:disable-next-line: max-line-length
"Select inputs are not yet supported. If you need to use one consider wrapping it in a Component which implements the ControlValueAccessor interface.");
var SelectInputControlDirective = /** @class */ (function (_super) {
    __extends(SelectInputControlDirective, _super);
    function SelectInputControlDirective(ref, r2, config) {
        var _this = _super.call(this, ref, r2, config) || this;
        throw SelectControlNotSupported;
        return _this;
    }
    SelectInputControlDirective.prototype.setValue = function (value) { };
    SelectInputControlDirective.prototype.getValue = function () {
        return null;
    };
    SelectInputControlDirective.ctorParameters = function () { return [
        { type: ElementRef },
        { type: Renderer2 },
        { type: undefined, decorators: [{ type: Inject, args: [CONFIG_TOKEN,] }] }
    ]; };
    SelectInputControlDirective = __decorate([
        Directive({
            selector: "select[" + CONTROL_DIRECTIVE_SELECTOR + "]",
        }),
        __param(2, Inject(CONFIG_TOKEN))
    ], SelectInputControlDirective);
    return SelectInputControlDirective;
}(AbstractControlDirective));

var NumberInputControlDirective = /** @class */ (function (_super) {
    __extends(NumberInputControlDirective, _super);
    function NumberInputControlDirective(ref, r2, config) {
        return _super.call(this, ref, r2, config) || this;
    }
    NumberInputControlDirective.prototype.onInput = function () {
        this.emitValue(this.getValue());
    };
    NumberInputControlDirective.prototype.onBlur = function () {
        this.emitTouched();
    };
    NumberInputControlDirective.prototype.getValue = function () {
        var value = this.ref.nativeElement.value;
        return value === '' ? null : parseFloat(value);
    };
    NumberInputControlDirective.prototype.setValue = function (value) {
        this.r2.setProperty(this.ref.nativeElement, 'value', value === null ? '' : value);
    };
    NumberInputControlDirective.prototype.setDisabled = function (disabled) {
        this.r2.setProperty(this.ref.nativeElement, 'disabled', disabled);
    };
    NumberInputControlDirective.ctorParameters = function () { return [
        { type: ElementRef },
        { type: Renderer2 },
        { type: undefined, decorators: [{ type: Inject, args: [CONFIG_TOKEN,] }] }
    ]; };
    __decorate([
        HostListener('input')
    ], NumberInputControlDirective.prototype, "onInput", null);
    __decorate([
        HostListener('blur')
    ], NumberInputControlDirective.prototype, "onBlur", null);
    NumberInputControlDirective = __decorate([
        Directive({
            selector: "input[type=\"number\"][" + CONTROL_DIRECTIVE_SELECTOR + "]",
        }),
        __param(2, Inject(CONFIG_TOKEN))
    ], NumberInputControlDirective);
    return NumberInputControlDirective;
}(AbstractControlDirective));

var TextInputControlDirective = /** @class */ (function (_super) {
    __extends(TextInputControlDirective, _super);
    function TextInputControlDirective(ref, r2, config) {
        return _super.call(this, ref, r2, config) || this;
    }
    TextInputControlDirective.prototype.onInput = function () {
        this.emitValue(this.getValue());
    };
    TextInputControlDirective.prototype.onBlur = function () {
        this.emitTouched();
    };
    TextInputControlDirective.prototype.getValue = function () {
        return this.ref.nativeElement.value;
    };
    TextInputControlDirective.prototype.setValue = function (value) {
        this.r2.setProperty(this.ref.nativeElement, 'value', value);
    };
    TextInputControlDirective.prototype.setDisabled = function (disabled) {
        this.r2.setProperty(this.ref.nativeElement, 'disabled', disabled);
    };
    TextInputControlDirective.ctorParameters = function () { return [
        { type: ElementRef },
        { type: Renderer2 },
        { type: undefined, decorators: [{ type: Inject, args: [CONFIG_TOKEN,] }] }
    ]; };
    __decorate([
        HostListener('input')
    ], TextInputControlDirective.prototype, "onInput", null);
    __decorate([
        HostListener('blur')
    ], TextInputControlDirective.prototype, "onBlur", null);
    TextInputControlDirective = __decorate([
        Directive({
            // tslint:disable-next-line: max-line-length
            selector: "input[type=\"text\"][" + CONTROL_DIRECTIVE_SELECTOR + "],textarea[" + CONTROL_DIRECTIVE_SELECTOR + "],input:not([type=\"checkbox\"]):not([type=\"number\"]):not([type=\"range\"]):not([type=\"radio\"])[" + CONTROL_DIRECTIVE_SELECTOR + "]",
        }),
        __param(2, Inject(CONFIG_TOKEN))
    ], TextInputControlDirective);
    return TextInputControlDirective;
}(AbstractControlDirective));

var ValueAccessorConnectorDirective = /** @class */ (function (_super) {
    __extends(ValueAccessorConnectorDirective, _super);
    function ValueAccessorConnectorDirective(ref, r2, config, accessors) {
        var _this = _super.call(this, ref, r2, config) || this;
        _this.initAccessor(accessors);
        return _this;
    }
    ValueAccessorConnectorDirective.prototype.initAccessor = function (accessors) {
        var _this = this;
        if (!accessors || !accessors.length) {
            return;
        }
        this.accessor = accessors[0];
        this.accessor.registerOnChange(function (val) { return _super.prototype.emitValue.call(_this, val); });
        this.accessor.registerOnTouched(function () { return _super.prototype.emitTouched.call(_this); });
    };
    ValueAccessorConnectorDirective.prototype.getValue = function () {
        return this.lastValue;
    };
    ValueAccessorConnectorDirective.prototype.setValue = function (value) {
        if (!this.accessor) {
            return;
        }
        this.accessor.writeValue(value);
        this.lastValue = value;
    };
    ValueAccessorConnectorDirective.prototype.setDisabled = function (disabled) {
        if (this.accessor && this.accessor.setDisabledState) {
            this.accessor.setDisabledState(disabled);
        }
    };
    ValueAccessorConnectorDirective.ctorParameters = function () { return [
        { type: ElementRef },
        { type: Renderer2 },
        { type: undefined, decorators: [{ type: Inject, args: [CONFIG_TOKEN,] }] },
        { type: Array, decorators: [{ type: Optional }, { type: Inject, args: [NG_VALUE_ACCESSOR,] }] }
    ]; };
    ValueAccessorConnectorDirective = __decorate([
        Directive({
            selector: "[" + CONTROL_DIRECTIVE_SELECTOR + "]",
        }),
        __param(2, Inject(CONFIG_TOKEN)),
        __param(3, Optional()),
        __param(3, Inject(NG_VALUE_ACCESSOR))
    ], ValueAccessorConnectorDirective);
    return ValueAccessorConnectorDirective;
}(AbstractControlDirective));

var CheckboxInputControlDirective = /** @class */ (function (_super) {
    __extends(CheckboxInputControlDirective, _super);
    function CheckboxInputControlDirective(ref, r2, config) {
        return _super.call(this, ref, r2, config) || this;
    }
    CheckboxInputControlDirective.prototype.onInput = function () {
        this.emitValue(this.getValue());
    };
    CheckboxInputControlDirective.prototype.onBlur = function () {
        this.emitTouched();
    };
    CheckboxInputControlDirective.prototype.getValue = function () {
        return this.ref.nativeElement.checked;
    };
    CheckboxInputControlDirective.prototype.setValue = function (value) {
        this.r2.setProperty(this.ref.nativeElement, 'checked', value);
    };
    CheckboxInputControlDirective.prototype.setDisabled = function (disabled) {
        this.r2.setProperty(this.ref.nativeElement, 'disabled', disabled);
    };
    CheckboxInputControlDirective.ctorParameters = function () { return [
        { type: ElementRef },
        { type: Renderer2 },
        { type: undefined, decorators: [{ type: Inject, args: [CONFIG_TOKEN,] }] }
    ]; };
    __decorate([
        HostListener('input')
    ], CheckboxInputControlDirective.prototype, "onInput", null);
    __decorate([
        HostListener('blur')
    ], CheckboxInputControlDirective.prototype, "onBlur", null);
    CheckboxInputControlDirective = __decorate([
        Directive({
            selector: "input[type=\"checkbox\"][" + CONTROL_DIRECTIVE_SELECTOR + "]",
        }),
        __param(2, Inject(CONFIG_TOKEN))
    ], CheckboxInputControlDirective);
    return CheckboxInputControlDirective;
}(AbstractControlDirective));

var controlDirectives = [
    TextInputControlDirective,
    CheckboxInputControlDirective,
    NumberInputControlDirective,
    RadioInputControlDirective,
    RangeInputControlDirective,
    SelectInputControlDirective,
    ValueAccessorConnectorDirective,
];

var ControlChildren = /** @class */ (function () {
    function ControlChildren() {
    }
    ControlChildren.prototype.getChildren = function () {
        var _this = this;
        var queries$ = [
            this.textInputs,
            this.numberInputs,
            this.radioInputs,
            this.rangeInputs,
            this.selectInputs,
            this.customInputs,
            this.checkboxInputs,
        ]
            .filter(Boolean)
            .map(function (query) {
            return _this.convertToObservable(query);
        });
        return combineLatest(queries$).pipe(map(function (queries) { return queries.reduce(function (q1, q2) { return __spread(q1, q2); }, []); }));
    };
    ControlChildren.prototype.convertToObservable = function (query) {
        return query.changes.pipe(map(function () { return query; }), startWith(query), map(function (list) { return list.toArray(); }));
    };
    __decorate([
        ContentChildren(TextInputControlDirective, { descendants: true })
    ], ControlChildren.prototype, "textInputs", void 0);
    __decorate([
        ContentChildren(CheckboxInputControlDirective, { descendants: true })
    ], ControlChildren.prototype, "checkboxInputs", void 0);
    __decorate([
        ContentChildren(NumberInputControlDirective, { descendants: true })
    ], ControlChildren.prototype, "numberInputs", void 0);
    __decorate([
        ContentChildren(RadioInputControlDirective, { descendants: true })
    ], ControlChildren.prototype, "radioInputs", void 0);
    __decorate([
        ContentChildren(RangeInputControlDirective, { descendants: true })
    ], ControlChildren.prototype, "rangeInputs", void 0);
    __decorate([
        ContentChildren(SelectInputControlDirective, { descendants: true })
    ], ControlChildren.prototype, "selectInputs", void 0);
    __decorate([
        ContentChildren(ValueAccessorConnectorDirective, { descendants: true })
    ], ControlChildren.prototype, "customInputs", void 0);
    return ControlChildren;
}());

var UnkownControlError = function (controlKey) {
    return new Error("The control \"" + controlKey + "\" is not part of the form.");
};
var AbstractFormDirective = /** @class */ (function (_super) {
    __extends(AbstractFormDirective, _super);
    function AbstractFormDirective(cdr) {
        var _this = _super.call(this) || this;
        _this.cdr = cdr;
        _this.formUpdate = new EventEmitter();
        _this.formSummary$ = new ReplaySubject(1);
        return _this;
    }
    Object.defineProperty(AbstractFormDirective.prototype, "setFormSummary", {
        set: function (summary) {
            if (!summary) {
                return;
            }
            this.formSummary$.next(summary);
        },
        enumerable: true,
        configurable: true
    });
    AbstractFormDirective.prototype.ngAfterViewInit = function () {
        var _this = this;
        var children$ = this.getChildren();
        children$
            .pipe(map(function (children) {
            return children.map(function (child) {
                return child.controlUpdate.pipe(map(function (update) { return [update, child.controlKey]; }));
            });
        }), switchMap(function (children) { return merge.apply(void 0, __spread(children)); }))
            .subscribe(function (_a) {
            var _b = __read(_a, 2), update = _b[0], key = _b[1];
            return _this.emitUpdate(update, key);
        });
        combineLatest([children$, this.formSummary$]).subscribe(function (_a) {
            var _b = __read(_a, 2), children = _b[0], summary = _b[1];
            return _this.updateChildren(children, summary);
        });
    };
    AbstractFormDirective.prototype.ngOnDestroy = function () {
        this.formSummary$.complete();
    };
    AbstractFormDirective.prototype.updateChildren = function (children, summary, shouldRetry) {
        if (shouldRetry === void 0) { shouldRetry = true; }
        return __awaiter(this, void 0, void 0, function () {
            var safeChildren, unknownControl, updatedChildren;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        safeChildren = children.filter(Boolean);
                        unknownControl = safeChildren.find(function (child) { return summary.controls[child.controlKey] === undefined; });
                        if (unknownControl && !shouldRetry) {
                            throw UnkownControlError(unknownControl.controlKey);
                        }
                        if (!unknownControl) return [3 /*break*/, 2];
                        this.cdr.detectChanges();
                        return [4 /*yield*/, this.getChildren().pipe(first()).toPromise()];
                    case 1:
                        updatedChildren = _a.sent();
                        return [2 /*return*/, this.updateChildren(updatedChildren, summary, false)];
                    case 2:
                        safeChildren.forEach(function (control) {
                            control.updateSummary(summary.controls[control.controlKey]);
                        });
                        return [2 /*return*/];
                }
            });
        });
    };
    AbstractFormDirective.ctorParameters = function () { return [
        { type: ChangeDetectorRef }
    ]; };
    __decorate([
        Input('formSummary')
    ], AbstractFormDirective.prototype, "setFormSummary", null);
    __decorate([
        Output()
    ], AbstractFormDirective.prototype, "formUpdate", void 0);
    AbstractFormDirective = __decorate([
        Directive()
    ], AbstractFormDirective);
    return AbstractFormDirective;
}(ControlChildren));

var FormGroupDirective = /** @class */ (function (_super) {
    __extends(FormGroupDirective, _super);
    function FormGroupDirective() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    FormGroupDirective.prototype.emitUpdate = function (update, key) {
        var _a;
        this.formUpdate.next({ controls: (_a = {}, _a[key] = update, _a) });
    };
    FormGroupDirective = __decorate([
        Directive({
            selector: '[ngrxFormGroup]',
        })
    ], FormGroupDirective);
    return FormGroupDirective;
}(AbstractFormDirective));

var FormArrayDirective = /** @class */ (function (_super) {
    __extends(FormArrayDirective, _super);
    function FormArrayDirective() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    FormArrayDirective.prototype.emitUpdate = function (update, key) {
        var index = parseInt(key.toString(), 10);
        var controls = new Array(index + 1);
        controls[index] = update;
        this.formUpdate.next({ controls: controls });
    };
    FormArrayDirective = __decorate([
        Directive({
            selector: '[ngrxFormArray]',
        })
    ], FormArrayDirective);
    return FormArrayDirective;
}(AbstractFormDirective));

var ɵ0 = defaultConfig;
/**
 * This module represents the entry point of ngrx-clean-forms.
 * It provides all necessary directives for the usage inside templates.
 *
 * The `.withConfig()` method can be used to pass a config to the module.
 *
 * - [npm](https://www.npmjs.com/package/ngrx-clean-forms)
 * - [documentation](https://github.com/NiklasPor/ngrx-clean-forms)
 */
var NgrxCleanFormsModule = /** @class */ (function () {
    function NgrxCleanFormsModule() {
    }
    NgrxCleanFormsModule_1 = NgrxCleanFormsModule;
    /**
     * Creates a new module with the given config. Missing values will be set with default values.
     *
     * @param config A partial `FormsConfig` object, supplying config paramters.
     *
     * @example
     * NgrxCleanFormsModule.withConfig({
     *   throttleTime: 15,
     * }),
     *
     */
    NgrxCleanFormsModule.withConfig = function (config) {
        return {
            ngModule: NgrxCleanFormsModule_1,
            providers: [
                {
                    provide: CONFIG_TOKEN,
                    useValue: __assign(__assign({}, defaultConfig), config),
                },
            ],
        };
    };
    var NgrxCleanFormsModule_1;
    NgrxCleanFormsModule = NgrxCleanFormsModule_1 = __decorate([
        NgModule({
            declarations: __spread([FormGroupDirective, FormArrayDirective], controlDirectives),
            imports: [CommonModule],
            exports: __spread([FormGroupDirective, FormArrayDirective], controlDirectives),
            providers: [
                {
                    provide: CONFIG_TOKEN,
                    useValue: ɵ0,
                },
            ],
        })
    ], NgrxCleanFormsModule);
    return NgrxCleanFormsModule;
}());

/**
 * Generated bundle index. Do not edit.
 */

export { NgrxCleanFormsModule, getFormArraySummary, getFormControlSummary, getFormGroupSummary, initFormArray, initFormControl, initFormGroup, mapFormGroupControlStates, mapFormGroupControlSummaries, mapFormGroupControlUpdates, reduceFormArray, reduceFormControl, reduceFormGroup, resetFormArray, resetFormControl, resetFormGroup, validatorOf, mapFormControls as ɵa, FormGroupDirective as ɵb, AbstractFormDirective as ɵc, ControlChildren as ɵd, TextInputControlDirective as ɵe, CONTROL_DIRECTIVE_SELECTOR as ɵf, AbstractControlDirective as ɵg, CONFIG_TOKEN as ɵh, defaultConfig as ɵi, CheckboxInputControlDirective as ɵj, NumberInputControlDirective as ɵk, RadioInputControlDirective as ɵl, RangeInputControlDirective as ɵm, SelectInputControlDirective as ɵn, ValueAccessorConnectorDirective as ɵo, FormArrayDirective as ɵp, controlDirectives as ɵq };
//# sourceMappingURL=ngrx-clean-forms.js.map
